
// Parser za microC sa semantickim akcijama
// Parser ne podrzava pozive funkcija kao argumente drugih poziva funkcija
%{
    #include <stdio.h>
    #include "defs.h"
    #include "symtab.h"
    #include "semantic.h"

    int yyparse(void);
    int yylex(void);
    int yyerror(char *s);
    void warning(char *s);

    extern int line;
    char char_buffer[CHAR_BUFFER_LENGTH];
    int level = GLOBAL_LEVEL;
    int var_num = 0;
    int arg_num = 0;
    int par_num = 0;
    int function_index = -1;
    int function_call_index = -1;
%}

%token _TYPE
%token _IF
%token _ELSE
%token _RETURN
%token _ID
%token _INT_NUMBER
%token _UNSIGNED_NUMBER
%token _LPAREN
%token _RPAREN
%token _COMMA
%token _LBRACKET
%token _RBRACKET
%token _ASSIGN
%token _SEMICOLON
%token _PLUS
%token _MINUS
%token _TIMES
%token _DIV
%token _RELOP


%%

program
    :   variable_list function_list
            {
                check_main();
            }

    |   function_list
            {
                check_main();
            }
    ;

variable_list
    :   variable _SEMICOLON
    |   variable_list variable _SEMICOLON
    ;

variable
    :   type _ID
            {
                switch(level) {
                  case GLOBAL_LEVEL : {
                      $$ = try_to_insert_id((char *)$2, GLOBAL_VAR, $1);
                  }; break;
                  case LOCAL_LEVEL : {
                      if (lookup_id((char *)$2, PARAMETER) == -1) {
                          $$ = try_to_insert_id((char *)$2, LOCAL_VAR, $1);
                          var_num++;
                          set_attribute($$, var_num);
                      }
                      else syyerror("redefinition of '%s'", (char *)$2);
                  }; break;
                  case PARAMETER_LEVEL : {
                      $$ = try_to_insert_id((char *)$2, PARAMETER, $1);
                      par_num++;
                      set_attribute($$, par_num);
                      set_param_type(function_index, par_num, $1);
                  }; break;
                }
            }
    ;

type
    :   _TYPE
        {
            $$ = $1;
        }
    ;

function_list
    :   function
    |   function_list function
    ;

function
    :   type _ID
            {
                if (lookup_id((char *)$2, GLOBAL_VAR) == -1)
                    function_index = try_to_insert_id((char *)$2, FUNCTION, $1);
                else
                    syyerror("redefinition of '%s'", (char *)$2);
            }
        _LPAREN
            {
                level = PARAMETER_LEVEL;
                par_num = 0;
            }
        parameters _RPAREN
            {
                // postavi broj parametara funkcije
                set_attribute(function_index, par_num);
                level = LOCAL_LEVEL;
                var_num = 0;
            }
        body
            {
                // izbaci iz tabele simbola sve lokalne simbole za funkciju
                if (get_last_element() > function_index)
                    clear_symbols(function_index + 1);
                level = GLOBAL_LEVEL;
            }
    ;

parameters
    :   /* empty */
    |   parameter_list
    ;

parameter_list
    :   variable
    |   parameter_list _COMMA variable
    ;

body
    :   _LBRACKET variable_list statement_list _RBRACKET
    |   _LBRACKET statement_list _RBRACKET
    ;

statement_list
    :   /* empty */
    |   statement_list statement
    ;

statement
    :   assignment_statement
    |   if_statement
    |   return_statement
    |   compound_statement
    ;

assignment_statement
    :   _ID
            {
                if( ($$ = lookup_id((char *)$1, GLOBAL_VAR|LOCAL_VAR|PARAMETER)) == -1 )
                    syyerror("invalid lvalue in assignment");
            }
        _ASSIGN num_exp _SEMICOLON
            {
                if(!check_types($2, $4))
                    syyerror("incompatible types in assignment");
            }
    ;

num_exp
    :   mul_exp

    |   num_exp _PLUS mul_exp
            {
                if(!check_types($1, $3))
                    syyerror("invalid operands to binary +");
            }

    |   num_exp _MINUS mul_exp
            {
                if(!check_types($1, $3))
                    syyerror("invalid operands to binary -");
            }
    ;

mul_exp
    :   exp

    |   mul_exp _TIMES exp
            {
                  if(!check_types($1, $3))
                      syyerror("invalid operands to binary *");
            }

    |   mul_exp _DIV exp
            {
                  if(!check_types($1, $3))
                      syyerror("invalid operands to binary /");
            }
    ;

exp
    :   constant

    |   _ID
            {
                if( ($$ = lookup_id((char *)$1, GLOBAL_VAR|LOCAL_VAR|PARAMETER)) == -1){
                    syyerror("'%s' undeclared", (char *)$1);
                }
            }

    |   function_call

    |   _LPAREN num_exp _RPAREN
            {
                $$ = $2;
            }
    ;

constant
    :   _INT_NUMBER
            {
                $$ = try_to_insert_constant((char *)$1, INT_TYPE);
            }

    |   _UNSIGNED_NUMBER
            {
                $$ = try_to_insert_constant((char *)$1, UNSIGNED_TYPE);
            }
    ;

function_call
    :   _ID
            {
                int index;
                if( (index = lookup_id((char *)$1, FUNCTION)) == -1 ) {
                    syyerror("'%s' is not a function", (char *)$1);
                }
                function_call_index = index;
            }
        _LPAREN arguments _RPAREN
            {
                check_arguments_number(function_call_index, arg_num);
                // postavi tip povratne vrednosti funkcije u %13
                set_register_type(FUNCTION_REGISTER, get_type(function_call_index));
                // povratna vrednost funkcije se uvek nalazi u %13
                $$ = FUNCTION_REGISTER;
            }
    ;

arguments
    :   /* empty */
            {
                arg_num = 0;
            }

    |   argument_list
    ;

argument_list
    :   num_exp
            {
                arg_num = 1;
                check_argument_type(function_call_index, $1, arg_num);
            }

    |   argument_list _COMMA num_exp
            {
                arg_num++;
                check_argument_type(function_call_index, $3, arg_num);
            }
    ;

if_statement
    :   if_part
    |   if_part _ELSE statement
    ;

if_part
    :   _IF _LPAREN rel_exp _RPAREN statement
    ;

rel_exp
    :   num_exp _RELOP num_exp
            {
                if(!check_types($1, $3))
                    syyerror("invalid operands to relational operator");
            }
    ;

return_statement
    :   _RETURN num_exp _SEMICOLON
            {
                if(!check_types(function_index, $2))
                    syyerror("incompatible types in return");
            }
    ;

compound_statement
    :   _LBRACKET statement_list _RBRACKET
    ;


%%

int yyerror(char *s) {
    fprintf(stderr, "\nERROR (%d): %s", line, s);
    return 0;
}

void warning(char *s) {
    fprintf(stderr, "\nWARNING (%d): %s", line, s);
}

int main() {
    printf("\nSTART\n");
    init_symtab();
    yyparse();
    clear_symtab();
    printf("\nSTOP\n");
    return 0;
}

