
// Parser za microC koji ispisuje drvo parsiranja

//  - indentacijom se obezbedjuje prikaz drveta parsiranja (izvrnuto)
//  - kada naidje na terminalni simbol doda 4 razmaka na indentaciju 
//    radi formiranja drveta parsiranja
//  - takodje stampa prazan red pre pojma kome prethodi racvanje

%{
    #include <stdio.h>
    #include <string.h>
    #include "defs.h"

    int yyparse(void);
    int yylex(void);
    int yyerror(char *s);
    void print_indent(void);
    void print(char *str);
    void print_quoted(char *str);
    void print_epsilon();

    extern int line;
    int indent = 0;

    static char* types[]  = { "int", "unsigned" };
    static char* relops[] = { "<", ">", "<=", ">=", "==", "!=" };
%}

%token _TYPE
%token _IF
%token _ELSE
%token _RETURN
%token _ID
%token _INT_NUMBER
%token _UNSIGNED_NUMBER
%token _LPAREN
%token _RPAREN
%token _COMMA
%token _LBRACKET
%token _RBRACKET
%token _ASSIGN
%token _SEMICOLON
%token _PLUS
%token _MINUS
%token _TIMES
%token _DIV
%token _RELOP


%%


program
    :   variable_list function_list
            {   print("program");   }

    |   function_list
            {	print("program");   }

    ;

variable_list
    :   variable _SEMICOLON
            {   print_quoted(";"); print("variable_list");   }

    |   variable_list variable _SEMICOLON
            {	print_quoted(";"); print("variable_list");   }
    ;

variable
    :   type _ID
            {	print_quoted((char *)$2); free((char *)$2); print("variable"); }
    ;

type
    :   _TYPE
            {	print_quoted(types[$1]); print("type");   }
    ;

function_list
    :   function
            {	print("function_list");   }

    |   function_list function
            {	print("function_list");   }
    ;

function
    :   type _ID
            {	print_quoted((char *)$2);  free((char *)$2); }
        _LPAREN 
            {	print_quoted("(");   }
        parameters _RPAREN 
            {	print_quoted(")");   }
        body
            {	print("function");   }
    ;

parameters
    :   /* empty */
            {	print_epsilon(); print("empty");   }

    |   parameter_list
            {	print("parameters");   }
    ;

parameter_list
    :   variable
            {	print("parameter_list");   }

    |   parameter_list _COMMA 
            {	print_quoted(",");   }
        variable
            {	print("parameter_list");   }
    ;

body
    :   _LBRACKET
            {	print_quoted("{");   }
        variable_list statement_list _RBRACKET
            {	print_quoted("}"); print("body");   }

    |   _LBRACKET 
            {	print_quoted("{");   }
        statement_list _RBRACKET
            {	print_quoted("}"); print("body");   }
    ;

statement_list
    :   /* empty */
            {	print_epsilon(); print("empty");   }

    |   statement_list statement
            {	print("statement_list");   }
    ;

statement
    :   assignment_statement
            {	print("statement");   }

    |   if_statement
            {	print("statement");   }

    |   return_statement
            {	print("statement");   }

    |   compound_statement
            {	print("statement");   }
    ;

assignment_statement
    :   _ID
            {	print_quoted((char *)$1); free((char *)$1); }
        _ASSIGN 
            {	print_quoted("=");   }
        num_exp _SEMICOLON
            {	print_quoted(";"); print("assignment_statement");   }
    ;

num_exp
    :   mul_exp
            {	print("num_exp");   }

    |   num_exp _PLUS 
            {	print_quoted("+");   }
        mul_exp
            {	print("num_exp");   }

    |   num_exp _MINUS 
            {	print_quoted("-");   }
        mul_exp
            {	print("num_exp");   }
    ;

mul_exp
    :   exp
            {	print("mul_exp");   }

    |   mul_exp _TIMES
            {	print_quoted("*");   }
        exp
            {	print("mul_exp");   }

    |   mul_exp _DIV
            {	print_quoted("/");   }
        exp
            {	print("mul_exp");   }
    ;

exp
    :   constant
            {	print("exp");   }

    |   _ID
            {	print_quoted((char *)$1); free((char *)$1); print("exp"); }

    |   function_call
            {	print("exp");   }

    |   _LPAREN 
            {	print_quoted("(");   }
        num_exp _RPAREN
            {	print_quoted(")"); print("exp");   }
    ;

constant
    :   _INT_NUMBER
            {	print_quoted((char *)$1); free((char *)$1); print("constant");   }

    |   _UNSIGNED_NUMBER
            {	print_quoted((char *)$1); free((char *)$1); print("constant");   }
    ;

function_call
    :   _ID
            {	print_quoted((char *)$1); free((char *)$1); }
        _LPAREN 
            {	print_quoted("(");   }
        arguments _RPAREN
            {	print_quoted(")"); print("function_call");   }
    ;

arguments
    :   /* empty */
            {	print_epsilon(); print("empty");   }

    |   argument_list
            {	print("arguments");   }
    ;

argument_list
    :   num_exp
            {	print("argument_list");   }

    |   argument_list _COMMA 
            {	print_quoted(",");   }
        num_exp
            {	print("argument_list");   }
    ;

if_statement
    :   if_part
            {	print("if_statement");   }

    |   if_part _ELSE 
            {	print_quoted("else");   }
        statement
            {	print("if_statement");   }
    ;

if_part
    :   _IF _LPAREN 
            {	print_quoted("if"); print_quoted("(");   }
        rel_exp _RPAREN 
            {	print_quoted(")");   }
        statement
            {	print("if_part");   }
    ;

rel_exp
    :   num_exp _RELOP 
            {	print_quoted(relops[$2]);   }
        num_exp
            {	print("rel_exp");   }
    ;

return_statement
    :   _RETURN 
            {	print_quoted("return");   }
        num_exp _SEMICOLON
            {	print_quoted(";"); print("return_statement");   }
    ;

compound_statement
    :   _LBRACKET 
            {	print_quoted("{");   }
        statement_list _RBRACKET
            {	print_quoted("}"); print("compound_statement");   }
    ;


%%

int yyerror(char *s) {
    fprintf(stderr, "\nERROR (%d): %s", line, s);
    return 0;
}

int main() {
    printf("\nSTART\n");
    yyparse();
    printf("\nSTOP\n");
    return 0;
}

void print_indent() {
    int i;
    printf("\n");
    for(i = 0; i < indent; i++)
        printf("    ");
}

void print(char *str) {
    printf("\n");
    print_indent();
    printf("%s", str);
}

void print_quoted(char *str) {
    indent++;
    print_indent();
    printf("\"%s\"", str);
}

void print_epsilon() {
    indent++;
    print_indent();
    printf("<epsilon>");
}
